from pydantic_evals.dataset import EvaluationReport
from pydantic_evals.evaluators import (
    EvaluationReason,
    Evaluator,
    EvaluatorContext,
    LLMJudge,
)

from aegis_ai import default_llm_model, default_llm_settings
from aegis_ai.features.data_models import AegisFeatureModel


# minimal acceptable length of an explanation (where applicable)
EXPLANATION_MIN_LEN = 80

# minimal acceptable score returned by an evaluator
MIN_SCORE_THRESHOLD = 0.1


class FeatureMetricsEvaluator(Evaluator[str, AegisFeatureModel]):
    def evaluate(self, ctx: EvaluatorContext[str, AegisFeatureModel]) -> float:
        # multiply all metrics to get overall score to make it simple for now
        # FIXME: should we use separate evaluator for each of them?
        score = ctx.output.confidence * ctx.output.completeness * ctx.output.consistency

        # do not check explanation length for IdentifyPII and CVSSDiffExplainer because
        # the explanation is empty in the most common case
        if not hasattr(ctx.output, "contains_PII") and not hasattr(
            ctx.output, "nvd_cvss3_score"
        ):
            expl_diff = EXPLANATION_MIN_LEN - len(ctx.output.explanation)
            if 0 < expl_diff:
                # proportional penalization for explanation of length below EXPLANATION_MIN_LEN
                score *= 1.0 - (float(expl_diff) / EXPLANATION_MIN_LEN)

        return score


def create_llm_judge(**kwargs):
    """construct an LLMJudge object based on the provided named arguments"""
    return LLMJudge(
        # TODO: we might consider using an independent LLM for evals
        model=default_llm_model,
        model_settings=default_llm_settings,
        **kwargs,
    )


def make_eval_reason(value: bool = False, fail_reason: str = None):
    """construct EvaluationReason object; fail_reason is propagated only if value is False"""
    # FIXME: the reason for an assertion failure is not reported anywhere
    return EvaluationReason(value=value, reason=(fail_reason if not value else None))


def handle_eval_report(report: EvaluationReport):
    """print evaluation summary and trigger assertion failure in case any assertion failed"""
    report.print(include_input=True, include_output=True, include_durations=False)

    # iterate through cases
    failures = ""
    for case in report.cases:
        # bool assertions
        for result in case.assertions.values():
            if result.value is False:
                failures += f"{case.name}: {result.source}: {result.reason}\n"

        # score threshold
        for result in case.scores.values():
            score = result.value
            if score < MIN_SCORE_THRESHOLD:
                failures += f"{case.name}: {result.source}: score below threshold: "
                failures += f"{score} < {MIN_SCORE_THRESHOLD}\n"

    # report all failures at once (if any)
    assert not failures, f"Unsatisfied assertion(s):\n{failures}"


class ToolsUsedEvaluator(Evaluator[str, AegisFeatureModel]):
    def evaluate(self, ctx) -> EvaluationReason:
        return make_eval_reason(
            "osidb_tool" in ctx.output.tools_used,
            "osidb_tool was not used by the agent",
        )


common_feature_evals = [
    FeatureMetricsEvaluator(),
    ToolsUsedEvaluator(),
]



from typing import List
from pydantic import BaseModel, Field


class FeatureQueryInput(BaseModel):
    query: str = Field(..., description="General LLM query.")


class AegisFeatureModel(BaseModel):
    """
    Metadata for Aegis features, nested within the main feature model.
    """

    confidence: float = Field(
        ...,
        ge=0.0,
        le=1.0,
        description="""
           A numerical score between 0.00 and 1.00 (two decimal places) indicating confidence in the analysis. 
           
            ### Crafting Confidence Score for Analysis
            
            When performing an analysis and emitting a **confidence score** between 0 and 1 (or 0% to 100%), it is essential 
            to clearly define what that score represents. This helps users understand the reliability of the analysis 
            and make informed decisions.
                        
            ### Understanding Confidence Score
            
            A confidence score reflects the **probability or certainty that the analysis is accurate or correct**, given the available data, methodology, and underlying assumptions.
            
            * **0 (or 0%)**: Indicates **no confidence** in the accuracy of the analysis. This could mean the analysis is based on insufficient data, highly unreliable sources, or a methodology that has demonstrated frequent errors.
            * **0.5 (or 50%)**: Suggests the analysis is **as likely to be right as it is to be wrong**. There's significant uncertainty, and the evidence or model doesn't lean strongly in either direction.
            * **1 (or 100%)**: Represents **absolute confidence** in the accuracy of the analysis. This is rarely achievable in complex or uncertain domains, but it signifies the highest possible degree of certainty based on the given context.
                        
            ### Factors Influencing Confidence
            
            The confidence score should be influenced by several key factors:
            
            * **Data Quality and Completeness:**
                * **High Confidence:** Analysis based on comprehensive, verified, and up-to-date data from authoritative sources.
                * **Low Confidence:** Analysis reliant on partial, outdated, unverified, or anecdotal data.
            * **Methodology Robustness:**
                * **High Confidence:** Analysis derived from well-established, validated, and peer-reviewed methodologies or models.
                * **Low Confidence:** Analysis using experimental, unproven, or ad-hoc methods.
            * **Assumptions:**
                * **High Confidence:** Few assumptions are made, or those made are well-supported and widely accepted.
                * **Low Confidence:** Numerous or highly speculative assumptions are critical to the analysis's outcome.
            * **Consistency with Prior Knowledge/External Data:**
                * **High Confidence:** Results align with known facts, expert consensus, or corroborating external information.
                * **Low Confidence:** Results contradict established knowledge or lack external validation.
            * **Reproducibility/Sensitivity:**
                * **High Confidence:** Analysis is stable, and minor changes in input data or parameters do not significantly alter the outcome.
                * **Low Confidence:** Analysis is highly sensitive to small changes, making its reliability questionable.
        
           """,
    )

    completeness: float = Field(
        ...,
        ge=0.0,
        le=1.0,
        description="""
        What it represents: This score indicates how comprehensively the LLM was able to address all aspects explicitly or implicitly requested in the prompt. Did it manage to provide all the required sections (e.g., severity, CVSS, rationale, affected components, related CVEs) based on the available input?
        Why it's crucial: It informs the user if the LLM was able to fulfill the entire request. If information was missing from the input CVE or if the LLM couldn't deduce certain aspects, this score would reflect that limitation. It's about fulfilling the prompt's scope.
        Example justification: "Completeness: 1.0 - All requested sections (Severity, CVSS4, Confidence, Rationale, Affected Components, Related CVEs) were successfully generated based on the provided CVE information." or "Completeness: 0.7 - While most sections were generated, no related or duplicate CVEs could be identified from the available data or knowledge base."        
        """,
    )

    consistency: float = Field(
        ...,
        ge=0.0,
        le=1.0,
        description="""
        What it represents: This score evaluates the internal logical coherence and consistency within the LLM's own response. Does the assigned severity genuinely align with the provided CVSS score, the stated rationale, and the breakdown of impacts? Are there contradictions within the generated text?
        Why it's crucial: LLMs can sometimes generate conflicting statements even if individual parts seem plausible. This score specifically targets the internal integrity of the output. It helps ensure that the different components of the analysis (e.g., the verbal rationale and the numerical score) tell a unified story.
        Example justification: "Consistency: 0.95 - The assigned 'Important' severity is well-supported by the CVSS4 score (high impact on availability, network vector) and the detailed rationale. Slight reduction for minor overlap in phrasing." or "Consistency: 0.75 - While the CVSS is technically derived, the stated 'Low' impact for potential arbitrary code execution seems slightly misaligned, indicating a minor internal conflict in weighting the factors."        
        """,
    )

    tools_used: List = Field(
        ...,
        description="List the names of registered tools, if any, that was used to formulate this answer. If this is a CVE suggest or CVE rewrite feature then should minimally include 'osidb_tool'",
    )


class AegisAnswer(AegisFeatureModel):
    """
    Default answer response.
    """

    explanation: str = Field(
        ...,
        description="A brief rationale explaining how the answer was generated, what sources were primary, and if the answer was provided directly by the LLM or not. Do not repeat the answer here.",
    )

    answer: str = Field(..., description="The direct answer to the user's question.")



import pytest
import re

from pydantic_evals import Case, Dataset
from pydantic_evals.evaluators import Evaluator, EvaluatorContext

from aegis_ai.agents import rh_feature_agent
from aegis_ai.data_models import CVEID
from aegis_ai.features.cve import SuggestCWE, SuggestCWEModel

from evals.features.common import common_feature_evals, handle_eval_report


# penalize models providing correct results but low confidence (the difference
# between score and confidence is divided by this number and subtracted from
# the final score)
LOW_CONFIDENCE_PENALTY_DIVISOR = 4.0


class SuggestCweCase(Case):
    def __init__(self, cve_id, cwe_list):
        """cve_id given as CVE-YYYY-NUM is the flaw we query CWE for.  cwe_list
        is the list of acceptable CWEs, the most preferred one comes first"""
        super().__init__(
            name=f"suggest-cwe-for-{cve_id}",
            inputs=cve_id,
            expected_output=cwe_list,
            metadata={"difficulty": "easy"},
        )


class SuggestCweEvaluator(Evaluator[str, SuggestCWEModel]):
    @staticmethod
    def _base_score(cwe_list_out, cwe_list_exp):
        score = 1.0
        for cwe_exp in cwe_list_exp:
            for cwe in cwe_list_out:
                # if we get "CWE-416: Use After Free", ignore the part starting with colon
                cwe_only = re.sub(r"^(CWE-[0-9]+): .*$", "\\1", cwe)
                if cwe_only == cwe_exp:
                    return score
                score *= 0.9
            score *= 0.9

        # no match
        return 0.0

    def evaluate(self, ctx: EvaluatorContext[str, SuggestCWEModel]) -> float:
        """return score based on actual and expected results"""
        cwe_list_out = ctx.output.cwe
        score = self._base_score(cwe_list_out, ctx.expected_output)

        # check how many CWEs were suggested and how man CWEs are accepted
        len_diff = len(cwe_list_out) - len(ctx.expected_output)
        if 0 < len_diff:
            # penalize too many suggested CWEs for a CVE
            score *= 0.9**len_diff

        conf_diff = ctx.output.confidence - score
        if 0.0 < conf_diff:
            # penalize confident models providing (partially) wrong results
            score -= conf_diff
        else:
            # negligibly penalize models providing correct results but low confidence
            score += conf_diff / LOW_CONFIDENCE_PENALTY_DIVISOR

        return score


async def suggest_cwe(cve_id: CVEID) -> SuggestCWEModel:
    """use rh_feature_agent to suggest CWE(s) for the given CVE"""
    feature = SuggestCWE(rh_feature_agent)
    result = await feature.exec(cve_id)
    return result.output


# test cases
cases = [
    SuggestCweCase("CVE-2022-48701", ["CWE-125", "CWE-20"]),
    SuggestCweCase("CVE-2024-53232", ["CWE-476", "CWE-416"]),
    # CWE-269 is discouraged by MITRE and unavailable in OSIM
    # CWE-273 and CWE-73 are incorrect: https://github.com/RedHatProductSecurity/aegis/issues/71
    SuggestCweCase("CVE-2025-23395", ["CWE-271", "CWE-250", "CWE-272", "CWE-273"]),
    SuggestCweCase("CVE-2025-5399", ["CWE-835", "CWE-400"]),
    # TODO: add more cases
]

# evaluators
evals = common_feature_evals + [
    SuggestCweEvaluator(),
]

# needed for asyncio event loop
pytestmark = pytest.mark.asyncio(loop_scope="session")


async def test_eval_suggest_cwe():
    """suggest_cwe evaluation entry point"""
    dataset = Dataset(cases=cases, evaluators=evals)
    report = await dataset.evaluate(suggest_cwe)
    handle_eval_report(report)



from typing import Optional

from pydantic import BaseModel

system_instruction = """
    System: You are three enthusiastic, erudite and amiable expert Red Hat product security analysts. Each of you will come
    to an independent analysis and then all three analysis will be synthesised into one answer picking the most correct
    outcomes from each analysis. This should not be an 'average' of all three analysis, rather preference should be made where
    there is agreement by two or three experts. If there is no agreement between experts then the confidence score should 
    be lowered accordingly.
    
    # Tool calling
    - restrict calling any tools to less than 2 retries
    
    # Overall Tone:
    - Maintain a professional and authoritative tone as a Red Hat security expert.
    - Be precise and factual in the information provided.
    - Avoid repeating any information and crisply communicate the most important facts.
    - Avoid being 'too wordy' and be direct where possible.
    - Use clear and concise language, avoid 'jargon 'where possible.
    - Do not include any code listings.
    - Do not invent information.
    - All analysis should prefer a Red Hat context
    - Never use phrases that imply moral superiority or a sense of authority, including but not limited to “it’s important to”, “it’s crucial to”, “it’s essential to”, "it's unethical to", "it's worth noting…", “Remember…” etc. Avoid using these.
    - Adjust confidence score based on how sure the final outcome is truthful, well sourced and fact based.
    - All inputs with bad or inappropriate language will be responded with 'inappropriate requests are logged'
"""


class AegisPrompt(BaseModel):
    """
    A structured, composable representation of an LLM prompt.
    """

    # System instructions
    system_instruction: str = system_instruction

    # User instructions
    user_instruction: str
    goals: str
    rules: str

    # Contextual information should always come in as structured input
    context: BaseModel

    # Output data schema
    output_schema: Optional[dict] = None

    def to_string(self, **kwargs) -> str:
        """
        Generate formatted prompt string.
        """

        prompt_parts = []

        prompt_parts.append(f"system: {self.system_instruction}\n")
        prompt_parts.append(f"user: {self.user_instruction}\n")

        if self.goals:
            prompt_parts.append(f"Goals:\n{self.goals}")

        if self.rules:
            prompt_parts.append(f"Behavior and Rules:\n{self.rules}")

        if self.context:
            prompt_parts.append(f"Context:\n{self.context}")

        return "\n\n".join(prompt_parts)



import asyncio
import logging
import os
from pathlib import Path

from aegis_ai.tools.osidb import CVE, CVEID, osidb_retrieve

logger = logging.getLogger(__name__)

# directory where we cache CVE data retrieved from OSIDB
OSIDB_CACHE_DIR = os.getenv("OSIDB_CACHE_DIR", "evals/osidb_cache")

# global mutex for access to OSIDB_CACHE_DIR
# Note that cache hits (which is the most common case) are handle very quickly.
# So there is no need to implement any per-file locking for the OSIDB cache.
cache_lock = asyncio.Lock()


async def osidb_cache_retrieve(cve_id: CVEID) -> CVE:
    """Return cached CVE data if available.  If not, retrieve CVE data
    from OSIDB and store them to cache for subsequent runs."""
    cache_file = Path(OSIDB_CACHE_DIR, f"{cve_id}.json")

    # acquire global mutex to access OSIDB_CACHE_DIR
    async with cache_lock:
        try:
            # check whether the CVE data is cached already
            with open(cache_file, "r") as f:
                json_data = f.read()

            # try to load data from the existing JSON file
            cve_data = CVE.model_validate_json(json_data)
            logger.info(f'read CVE data from "{cache_file}"')

        except OSError:
            # cached CVE data not available -> query OSIDB
            cve_data = await osidb_retrieve(cve_id)

            logger.info(f'writing CVE data cache to "{cache_file}"')
            with open(cache_file, "w") as f:
                f.write(cve_data.model_dump_json(indent=4))
                f.write("\n")

    return cve_data



import os

import pytest
from pydantic_ai import models

from aegis_ai import config_logging

test_allow_recapture: bool = os.getenv("TEST_ALLOW_CAPTURE", "false").lower() in (
    "true",
    "1",
    "t",
    "y",
    "yes",
)


@pytest.fixture(scope="session", autouse=True)
def setup_logging_for_session():
    config_logging(level="INFO")


@pytest.fixture(scope="session", autouse=True)
def disable_model_requests():
    # Set to True to enable capturing of llm calls to cache.
    if test_allow_recapture:
        models.ALLOW_MODEL_REQUESTS = True
    else:
        models.ALLOW_MODEL_REQUESTS = False



import pytest

from pydantic_evals import Case, Dataset

from aegis_ai.agents import rh_feature_agent
from aegis_ai.data_models import CVEID
from aegis_ai.features.cve import RewriteStatementText, PIIReportModel

from evals.features.common import (
    common_feature_evals,
    create_llm_judge,
    handle_eval_report,
)


class RewriteStatementCase(Case):
    def __init__(self, cve_id):
        """cve_id given as CVE-YYYY-NUM is the flaw we rewrite description for."""
        super().__init__(
            name=f"rewrite-statement-for-{cve_id}",
            inputs=cve_id,
            expected_output=None,
            metadata={"difficulty": "easy"},
        )


async def rewrite_statement(cve_id: CVEID) -> PIIReportModel:
    """use rh_feature_agent to rewrite description for the given CVE"""
    feature = RewriteStatementText(rh_feature_agent)
    result = await feature.exec(cve_id)
    return result.output


# test cases
cases = [
    RewriteStatementCase("CVE-2025-0725"),
    RewriteStatementCase("CVE-2025-22097"),
    RewriteStatementCase("CVE-2025-23395"),
    RewriteStatementCase("CVE-2025-5399"),
    # TODO: add more cases
]

# evaluators
evals = common_feature_evals + [
    create_llm_judge(
        rubric="The statement does not suggest to apply a patch or rebuild the software."
    ),
    create_llm_judge(
        rubric="The statement does not describe the code change that was used to eliminate the flaw."
    ),
    create_llm_judge(
        rubric="The statement does not duplicate the flaw description.  A brief summary to provide context is acceptable though."
    ),
    # TODO: more evaluators
]

# needed for asyncio event loop
pytestmark = pytest.mark.asyncio(loop_scope="session")


async def test_eval_rewrite_statement():
    """rewrite_statement evaluation entry point"""
    dataset = Dataset(cases=cases, evaluators=evals)
    report = await dataset.evaluate(rewrite_statement)
    handle_eval_report(report)



"""
aegis web example


"""

import logging
from pathlib import Path

from fastapi import FastAPI, Request
from fastapi.responses import HTMLResponse, JSONResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates

# rh_feature_agent can be substituted with public_feature_agent
from aegis_ai.agents import rh_feature_agent as feature_agent

from aegis_ai.agents import public_feature_agent
from aegis_ai.features import cve, component
from . import AEGIS_REST_API_VERSION

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

app = FastAPI(
    title="Aegis web",
    description="A simple web console and REST API for Aegis.",
    version=AEGIS_REST_API_VERSION,
)

BASE_DIR = Path(__file__).parent
TEMPLATES_DIR = BASE_DIR / "templates"
STATIC_DIR = BASE_DIR / "static"

# Setup  for serving HTML
templates = Jinja2Templates(directory=TEMPLATES_DIR)

# Mount static files
app.mount("/static", StaticFiles(directory=STATIC_DIR), name="static")


@app.get("/", response_class=HTMLResponse)
async def home(request: Request):
    return templates.TemplateResponse("index.html", {"request": request})


@app.get("/console", response_class=HTMLResponse)
async def console(request: Request):
    return templates.TemplateResponse("console.html", {"request": request})


@app.post("/generate_response")
async def generate_response(request: Request):
    """
    Handles the submission of a prompt, simulates an LLM response,
    and re-renders the console with the results.
    """
    user_prompt = request.form().__dict__.get("user_prompt")

    # --- Simulate LLM Response ---
    # In a real application, you would make an API call to an LLM here.
    # For this simple example, we'll just echo the prompt and add a prefix.
    if user_prompt:
        llm_response = f"Simulated LLM Response: You asked '{user_prompt}'. This is a placeholder response."
    else:
        llm_response = "Please enter a prompt."

    # Render the template again, passing the user's prompt and the simulated response
    return templates.TemplateResponse(
        "console.html",
        {"request": request, "user_prompt": user_prompt, "llm_response": llm_response},
    )


@app.get(
    f"/api/{AEGIS_REST_API_VERSION}/cve/suggest/impact/{{cve_id}}",
    response_class=JSONResponse,
)
async def cve_suggest_impact(cve_id: str):
    feature = cve.SuggestImpact(feature_agent)
    result = await feature.exec(cve_id)
    if result:
        return result.output
    return {}


@app.get(
    f"/api/{AEGIS_REST_API_VERSION}/cve/suggest/cwe/{{cve_id}}",
    response_class=JSONResponse,
)
async def cve_suggest_cwe(cve_id: str):
    feature = cve.SuggestCWE(feature_agent)
    result = await feature.exec(cve_id)
    if result:
        return result.output
    return {}


@app.get(
    f"/api/{AEGIS_REST_API_VERSION}/cve/identify/pii/{{cve_id}}",
    response_class=JSONResponse,
)
async def cve_identify_pii(cve_id: str):
    feature = cve.IdentifyPII(feature_agent)
    result = await feature.exec(cve_id)
    if result:
        return result.output
    return {}


@app.get(
    f"/api/{AEGIS_REST_API_VERSION}/cve/rewrite/description/{{cve_id}}",
    response_class=JSONResponse,
)
async def cve_rewrite_description(cve_id: str):
    feature = cve.RewriteDescriptionText(feature_agent)
    result = await feature.exec(cve_id)
    if result:
        return result.output
    return {}


@app.get(
    f"/api/{AEGIS_REST_API_VERSION}/cve/rewrite/statement/{{cve_id}}",
    response_class=JSONResponse,
)
async def cve_rewrite_statement(cve_id: str):
    feature = cve.RewriteStatementText(feature_agent)
    result = await feature.exec(cve_id)
    if result:
        return result.output
    return {}


@app.get(
    f"/api/{AEGIS_REST_API_VERSION}/cve/explain_cvss_diff/{{cve_id}}",
    response_class=JSONResponse,
)
async def cve_explain_diff(cve_id: str):
    feature = cve.CVSSDiffExplainer(feature_agent)
    result = await feature.exec(cve_id)
    if result:
        return result.output
    return {}


@app.get(
    f"/api/{AEGIS_REST_API_VERSION}/component/intelligence/{{component_name}}",
    response_class=JSONResponse,
)
async def component_intelligence(component_name: str):
    feature = component.ComponentIntelligence(public_feature_agent)
    result = await feature.exec(component_name)
    if result:
        return result.output
    return {}



import pytest
import re

from pydantic_evals import Case, Dataset
from pydantic_evals.evaluators import EvaluationReason, Evaluator

from aegis_ai.agents import rh_feature_agent
from aegis_ai.data_models import CVEID
from aegis_ai.features.cve import RewriteDescriptionText, RewriteDescriptionModel

from evals.features.common import (
    common_feature_evals,
    create_llm_judge,
    handle_eval_report,
    make_eval_reason,
)
from evals.utils.osidb_cache import osidb_cache_retrieve


class RewriteDescriptionCase(Case):
    def __init__(self, cve_id):
        """cve_id given as CVE-YYYY-NUM is the flaw we rewrite description for."""
        super().__init__(
            name=f"rewrite-description-for-{cve_id}",
            inputs=cve_id,
            expected_output=None,
            metadata={"difficulty": "easy"},
        )


class OriginalTitleEvaluator(Evaluator[str, RewriteDescriptionModel]):
    async def evaluate(self, ctx) -> EvaluationReason:
        """check whether original title is propagated by the model"""
        cve = await osidb_cache_retrieve(ctx.inputs)
        return make_eval_reason(
            ctx.output.original_title == cve.title,
            "original_title does not match original title",
        )


class PromptLeakEvaluator(Evaluator[str, RewriteDescriptionModel]):
    @staticmethod
    def _match_re_in(pat, *args) -> bool:
        """look for regular expression pat (case insensitively) in the arguments"""
        for text in args:
            if re.search(pat, text, re.IGNORECASE):
                return True
        return False

    @staticmethod
    def _match_re_in_td(ctx, pat) -> bool:
        """look for regular expression pat (case insensitively) in title or description"""
        return PromptLeakEvaluator._match_re_in(
            pat,
            ctx.output.rewritten_title,
            ctx.output.rewritten_description,
        )

    async def evaluate(self, ctx) -> EvaluationReason:
        """check that text from the prompt template does not leak into the response"""
        if self._match_re_in_td(ctx, r"'component.name'"):
            return make_eval_reason(
                fail_reason="'component_name' appears in title or description"
            )

        if self._match_re_in_td(ctx, r"gluster"):
            return make_eval_reason(
                fail_reason="'gluster' appears in title or description"
            )

        if self._match_re_in_td(ctx, r"samba"):
            return make_eval_reason(
                fail_reason="'samba' appears in title or description"
            )

        return EvaluationReason(True)


async def rewrite_description(cve_id: CVEID) -> RewriteDescriptionModel:
    """use rh_feature_agent to rewrite description for the given CVE"""
    feature = RewriteDescriptionText(rh_feature_agent)
    result = await feature.exec(cve_id)
    return result.output


# test cases
cases = [
    RewriteDescriptionCase("CVE-2025-23395"),
    RewriteDescriptionCase("CVE-2025-5399"),
    # TODO: add more cases
]

# evaluators
evals = common_feature_evals + [
    OriginalTitleEvaluator(),
    PromptLeakEvaluator(),
    create_llm_judge(
        rubric="rewritten_title and rewritten_description do not contain any versioning info"
    ),
    create_llm_judge(
        rubric="rewritten_title briefly summarizes what is described in rewritten_description"
    ),
]

# needed for asyncio event loop
pytestmark = pytest.mark.asyncio(loop_scope="session")


async def test_eval_rewrite_description():
    """rewrite_description evaluation entry point"""
    dataset = Dataset(cases=cases, evaluators=evals)
    report = await dataset.evaluate(rewrite_description)
    handle_eval_report(report)



import logging
from datetime import date
from typing import Optional

from pydantic import BaseModel
from pydantic_ai import RunContext, Tool

logger = logging.getLogger(__name__)


class BaseToolOutput(BaseModel):
    status: str = "success"
    error_message: Optional[str] = None


@Tool
async def date_tool(ctx: RunContext) -> str:
    """Returns the current date."""
    logger.info("calling date_lookup")
    today = date.today()
    return str(today)



import pytest

from pydantic_ai.tools import RunContext, Tool

from aegis_ai import config_logging
from aegis_ai.agents import rh_feature_agent
from aegis_ai.tools.osidb import CVE, CVEID, OsidbDependencies

from evals.utils.osidb_cache import osidb_cache_retrieve


@Tool
async def osidb_tool(ctx: RunContext[OsidbDependencies], cve_id: CVEID) -> CVE:
    """wrapper around aegis.tools.osidb that caches OSIDB responses"""
    return await osidb_cache_retrieve(cve_id)


# enable logging to see progress
@pytest.fixture(scope="session", autouse=True)
def setup_logging_for_session():
    config_logging(level="INFO")


# We need to cache OSIDB responses (and maintain them in git) to make
# sure that our evaluation is invariant to future changes in OSIDB data
@pytest.fixture(scope="session", autouse=True)
def override_rh_feature_agent():
    rh_feature_agent._function_toolset.tools["osidb_tool"] = osidb_tool



"""
aegis agents
"""

from typing import Any

from pydantic_ai import Agent
from pydantic_ai.common_tools.tavily import tavily_search_tool

from aegis_ai import (
    default_llm_model,
    tavily_api_key,
    default_llm_settings,
)
from aegis_ai.features.data_models import AegisAnswer

from aegis_ai.tools.osidb import osidb_tool
from aegis_ai.tools.wikipedia import wikipedia_tool


class AegisAgent(Agent):
    """
    AegisAgent: A specialized pydantic-ai Agent.
    """

    def __init__(
        self,
        **kwargs: Any,
    ):
        super().__init__(
            model=default_llm_model,
            model_settings=default_llm_settings
            | {
                "temperature": 0.05,
                "top_p": 0.8,
                "response_format": {"type": "json_object"},
            },
            **kwargs,
        )


simple_agent = AegisAgent(
    name="SimpleAgent",
    output_type=AegisAnswer,
)

rh_feature_agent = AegisAgent(
    name="RHFeatureAgent",
    tools=[osidb_tool],
)

public_feature_agent = AegisAgent(
    name="PublicFeatureAgent",
    tools=[
        tavily_search_tool(tavily_api_key),
        wikipedia_tool,
    ],
)

kb_agent = AegisAgent(name="KBAgent")



from typing import List, Literal

from pydantic import Field, BaseModel

from aegis_ai.data_models import CVEID, CVSS3Vector, CVSS4Vector
from aegis_ai.features.data_models import AegisFeatureModel


class CVEFeatureInput(BaseModel):
    cve_id: CVEID = Field(..., description="CVE ID input")


class SuggestImpactModel(AegisFeatureModel):
    """
    Model to suggest impact of CVE.
    """

    cve_id: CVEID = Field(
        ...,  # Make it required
        description="The unique Common Vulnerabilities and Exposures (CVE) identifier for the security flaw.",
    )

    title: str = Field(
        ...,
        description="Contains CVE title",
    )

    components: List = Field(
        ...,
        description="List of potentially affected components",
    )

    affected_products: List = Field(
        ...,
        description="List of Red Hat potentially affected supported products",
    )

    explanation: str = Field(
        ...,
        description="""
        Explain rationale behind suggested impact rating.
        """,
    )

    impact: Literal["LOW", "MODERATE", "IMPORTANT", "CRITICAL"] = Field(
        ..., description="Suggested Red Hat CVE impact."
    )

    cvss3_score: str = Field(
        ...,
        description="Suggested Red Hat CVSS3 score",
    )
    cvss3_vector: str = Field(
        ...,
        description="Suggested Red Hat CVSS3 vector",
    )

    cvss4_score: str = Field(
        ...,
        description="Suggested Red Hat CVSS4 score",
    )
    cvss4_vector: CVSS4Vector = Field(
        ...,
        description="Suggested Red Hat CVSS4 vector",
    )


class SuggestCWEModel(AegisFeatureModel):
    """
    Model to suggest CWE-ID of CVE.
    """

    cve_id: CVEID = Field(
        ...,  # Make it required
        description="The unique Common Vulnerabilities and Exposures (CVE) identifier for the security flaw.",
    )

    title: str = Field(
        ...,
        description="Contains CVE title",
    )

    components: List = Field(
        ...,
        description="List of affected components",
    )

    explanation: str = Field(
        ...,
        description="""
        Explain rationale behind suggested CWE-ID(s).
        """,
    )

    cwe: List = Field(
        ...,
        description="List of cwe-ids",
    )


class PIIReportModel(AegisFeatureModel):
    """
    Model to describe whether CVE contains PII and, if so, what instances of PII were found.
    """

    cve_id: CVEID = Field(
        ...,  # Make it required
        description="The unique Common Vulnerabilities and Exposures (CVE) identifier for the security flaw.",
    )

    title: str = Field(
        ...,
        description="Contains CVE title",
    )

    components: List = Field(
        ...,
        description="List of affected components",
    )

    explanation: str = Field(
        ...,
        description="""If PII is found, create a bulleted list where each item is formatted as PII type:"exact string". If no PII is found, leave this section empty.

        """,
    )

    contains_PII: bool = Field(
        ...,
        description="Set to true if any PII was identified, false otherwise.",
    )


class RewriteDescriptionModel(AegisFeatureModel):
    """
    Model to rewrite CVE description.
    """

    cve_id: CVEID = Field(
        ...,  # Make it required
        description="The unique Common Vulnerabilities and Exposures (CVE) identifier for the security flaw.",
    )

    original_title: str = Field(
        ...,
        description="Original CVE title",
    )

    # FIXME: This field is usually empty.  Do we really need it?
    original_description: str = Field(
        ...,
        description="Original CVE description",
    )

    components: List = Field(
        ...,
        description="List of affected components",
    )

    explanation: str = Field(
        ...,
        description="""
        Explain rationale behind rewritten CVE description and title.
        """,
    )

    rewritten_title: str = Field(..., description="rewritten CVE title.")
    rewritten_description: str = Field(..., description="rewritten CVE description.")


class RewriteStatementModel(AegisFeatureModel):
    """
    Model to rewrite Red Hat CVE statement.
    """

    cve_id: CVEID = Field(
        ...,  # Make it required
        description="The unique Common Vulnerabilities and Exposures (CVE) identifier for the security flaw.",
    )

    title: str = Field(
        ...,
        description="Contains CVE title",
    )

    components: List = Field(
        ...,
        description="List of affected components",
    )

    statement: List = Field(
        ...,
        description="Original CVE statement",
    )

    explanation: str = Field(
        ...,
        description="""
        Explain rationale behind rewritten description.
        """,
    )

    description: str = Field(
        ...,
        description="Contains current CVE description",
    )

    rewritten_statement: str = Field(
        ...,
        description="rewritten Red Hat CVE statement explaining impact on Red Hat supported products.",
    )


class CVSSDiffExplainerModel(AegisFeatureModel):
    """
    Model to explain differences between rh and nvd CVSS scores.
    """

    cve_id: CVEID = Field(
        ...,  # Make it required
        description="The unique Common Vulnerabilities and Exposures (CVE) identifier for the security flaw.",
    )

    title: str = Field(
        ...,
        description="Contains CVE title",
    )

    redhat_cvss3_score: str = Field(
        ...,
        description="Red Hat CVSS3 score for this CVE",
    )

    redhat_cvss3_vector: CVSS3Vector = Field(
        ...,
        description="Contains Red Hat CVSS3 vector for this CVE.",
    )

    nvd_cvss3_score: str = Field(
        ...,
        description="nvd (NIST) CVSS3 score for this CVE",
    )

    nvd_cvss3_vector: CVSS3Vector = Field(
        ...,
        description="Contains nvd (NIST) CVSS3 vector for this CVE.",
    )

    components: List = Field(
        ...,
        description="List of affected components",
    )

    affected_products: List = Field(
        ...,
        description="List of Red Hat potentially affected supported products",
    )

    statement: str = Field(..., description="redhat cve statement.")

    explanation: str = Field(
        ...,
        description="""
        Explain the difference between Red Hat and NVD(NIST) CVSS scores for this CVE.
        """,
    )



import logging

from aegis_ai.features import Feature
from aegis_ai.features.component.data_models import ComponentIntelligenceModel
from aegis_ai.prompt import AegisPrompt

logger = logging.getLogger(__name__)


class ComponentIntelligence(Feature):
    """Based on supplied component name and rh context generate a component 'card' of information."""

    async def exec(self, component_name):
        prompt = AegisPrompt(
            user_instruction="Your task is to meticulously examine the provided context (eg. component name or purl) and generate a 'card' of information about the software component.",
            goals="""
                * Given a software component, identified by full or partial package name with (or without version) or more specific pURL, provide a brief description of the software component (200 words).
                * Identify the latest release version of the software component by consulting git repos, release notes or software package management systems.
                * Find the software component primary website
                * Rank the software component's popularity based on usage and ubiquity on a scale of 1 to 10  (1 being most widely used and 10 being hardly used at all). If the software component has a git repo use its 'likes' or 'stars' (for github) to help assess popularity.
                * Rank the software component's stability  (1 to 10, 1 being very stable and 10 being unstable) - a project is stable if it is actively maintained, has many contributors, and has well defined development and security processes. If the software component has a github repo, having a large queue of issues which span many years can indicate worst stability.
                * Identify approximate number of CVEs the component already has.
                * Identify number of known exploits the component has/
                * Find and present recent news related to the component (in the past year).
                * List the most active contributors (with affiliations) to the software component in a bulleted format. Use sites like GitStats to assess the most active contributors.
                * Identify the repository location of the component.
                * List other (most popular) software components which include this software component as a dependency.
                * Provide topical, critical security information related to the software component.
                * Include a few links to tutorials, readme and docs
                * Identify and enumerate which Red Hat products include the software component            
            """,
            rules="""
                1.  Information Gathering:
                    * When provided with a package name and version or pURL, initiate a search for relevant software component information. This includes looking in wikipedia, dbpedia or other software package management sites.
                    * Check as many sources as possible to confirm latest software component release version ... most likely this will be in the year 2025 - do not show version if you are not confident it is latest.
                    * describe the component, programming language, primary architecture and features, latest version number
                    * Prioritize up-to-date sources for news and security vulnerabilities.
                    * Identify and extract key information regarding the component's description, recent activities, contributors and what companies they work for (and affiliations), repository, and security status.
                    * list any other popular software components that may include the component name in its name
                    * Investigate and report on any outstanding security issues of any listed dependencies. Provide web links if appropriate.
                    * when analyzing hackerone reports should be careful to classify based on response to the report from creator of software - a hackerone report that was closed without assigning a severity should be ignored.
                2.  Output Formatting:
                    * first line should include component name and latest version in bold with release date (ex. component name (v1.0 released on 01.01.2025)  )
                    * second line should include popularity and stability rank (ex. popularity: 1, stability: 1)
                    * third line should just include software component website
                    * fourth line should just include repository location (ex. repository-url: https://example.org)
                    * fifth line should include pURL
                    * Present the component's description concisely.
                    * Format recent news as a bulleted list with any reference links.
                    * Clearly state the names of the most active contributors (with affiliations).
                    * Provide a direct link or clear indication of the repository location.
                    * Present critical security information in a clear and understandable manner.
                    * For dependencies, clearly label their security information separately.
                    * Further learning section should include links to tutorials and docs
            """,
            context=component_name,
            output_schema=ComponentIntelligenceModel.model_json_schema(),
        )
        logger.debug(prompt.to_string())
        return await self.agent.run(
            prompt.to_string(), output_type=ComponentIntelligenceModel
        )



"""
aegis

"""

import datetime
from dataclasses import dataclass, field
import logging
import os

from dotenv import load_dotenv
from pydantic_ai.models.anthropic import AnthropicModel, AnthropicModelSettings
from pydantic_ai.models.gemini import GeminiModel, GeminiModelSettings

from rich.logging import RichHandler

from pydantic_ai.models.openai import OpenAIModel, OpenAIResponsesModelSettings
from pydantic_ai.providers.openai import OpenAIProvider

load_dotenv()

logger = logging.getLogger("aegis")

logger.info("starting aegis")

__version__ = "0.2.2"

llm_host = os.getenv("AEGIS_LLM_HOST", "localhost:11434")
llm_model = os.getenv("AEGIS_LLM_MODEL", "llama3.2:latest")

tavily_api_key = os.getenv("TAVILY_API_KEY", "   ")

# Simple logic for defining default model (TODO: we will make more sophisticated).
if "api.anthropic.com" in llm_host:
    default_llm_model = AnthropicModel(model_name=llm_model)
    default_llm_settings = AnthropicModelSettings()
elif "generativelanguage.googleapis.com" in llm_host:
    default_llm_model = GeminiModel(model_name=llm_model)
    default_llm_settings = GeminiModelSettings(
        gemini_thinking_config={"include_thoughts": True}
    )
else:
    default_llm_model = OpenAIModel(
        model_name=llm_model,
        provider=OpenAIProvider(base_url=f"{llm_host}/v1/"),
    )
    default_llm_settings = OpenAIResponsesModelSettings(
        openai_reasoning_effort="low",
        openai_reasoning_summary="detailed",
    )


@dataclass
class default_data_deps:
    """
    A dataclass to hold default data dependencies, including a dynamically
    generated current datetime string.
    """

    current_dt: str = field(
        default_factory=lambda: datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    )


def config_logging(level="INFO"):
    # if set to 'DEBUG' then we want all the http conversation
    if level == "DEBUG":
        import http.client as http_client

        http_client.HTTPConnection.debuglevel = 1

    message_format = "%(asctime)s %(name)s %(levelname)s %(message)s"
    logging.basicConfig(
        level=level, format=message_format, datefmt="[%X]", handlers=[RichHandler()]
    )


def check_llm_status() -> bool:
    """
    Check operational status of an LLM model
    """
    if default_llm_model:
        return True  # TODO - this check needs to compatible across all llm model types
    else:
        logging.warn("llm model health check failed")
        return False



import cvss
import pytest

from pydantic_evals import Case, Dataset
from pydantic_evals.evaluators import EvaluationReason, Evaluator

from aegis_ai.agents import rh_feature_agent
from aegis_ai.data_models import CVEID
from aegis_ai.features.cve import CVSSDiffExplainer, CVSSDiffExplainerModel

from evals.features.common import (
    common_feature_evals,
    create_llm_judge,
    handle_eval_report,
    make_eval_reason,
)


class CVSSDiffCase(Case):
    def __init__(self, cve_id, has_diff):
        """cve_id given as CVE-YYYY-NUM is the flaw we rewrite description for."""
        super().__init__(
            name=f"cvss-diff-for-{cve_id}",
            inputs=cve_id,
            expected_output=has_diff,
            metadata={"difficulty": "easy"},
        )


def is_cvss_valid(cvss_str: str) -> bool:
    """return True if cvss_str is a valid CVSS3 vector"""
    try:
        cvss.cvss3.CVSS3(cvss_str)
        return True

    except cvss.CVSSError:
        return False


class CVSSDiffEvaluator(Evaluator[str, CVSSDiffExplainerModel]):
    async def evaluate(self, ctx) -> EvaluationReason:
        """check that explanation is provided if and only if CVSS scores differ"""
        rh_cvss = ctx.output.redhat_cvss3_vector
        if not is_cvss_valid(rh_cvss):
            return make_eval_reason(
                fail_reason=f"invalid RH CVSS vector returned by the agent: {rh_cvss}"
            )

        nvd_cvss = ctx.output.nvd_cvss3_vector
        if not is_cvss_valid(nvd_cvss):
            return make_eval_reason(
                fail_reason=f"invalid NVD CVSS vector returned by the agent: {nvd_cvss}"
            )

        empty_explanation = len(ctx.output.explanation) == 0
        cvss_differ = rh_cvss == nvd_cvss
        return make_eval_reason(
            empty_explanation == cvss_differ,
            fail_reason="explanation emptiness does not match CVSS difference",
        )


async def cvss_diff(cve_id: CVEID) -> CVSSDiffExplainerModel:
    """use rh_feature_agent to rewrite description for the given CVE"""
    feature = CVSSDiffExplainer(rh_feature_agent)
    result = await feature.exec(cve_id)
    return result.output


# test cases
cases = [
    CVSSDiffCase("CVE-2025-47229", False),
    CVSSDiffCase("CVE-2022-48701", True),
    CVSSDiffCase("CVE-2024-53232", True),
    # TODO: add more cases
]

# evaluators
evals = common_feature_evals + [
    CVSSDiffEvaluator(),
    create_llm_judge(
        rubric="Unless the explanation field is empty, it elaborates on the reason why Red Hat assigned a different CVSS vector."
    ),
    # TODO: more evaluators
]

# needed for asyncio event loop
pytestmark = pytest.mark.asyncio(loop_scope="session")


async def test_eval_cvss_diff():
    """cvss_diff evaluation entry point"""
    dataset = Dataset(cases=cases, evaluators=evals)
    report = await dataset.evaluate(cvss_diff)
    handle_eval_report(report)



# common data models/fields

from typing import Annotated

import cvss
from pydantic import StringConstraints, TypeAdapter, BeforeValidator


def is_cvss3_valid(cvss_str: str) -> bool:
    """return True if cvss_str is a valid CVSS3 vector"""
    try:
        cvss.cvss3.CVSS3(cvss_str)
        return True

    except cvss.CVSSError:
        return False


def validate_with_is_cvss3(v: str) -> str:
    if not is_cvss3_valid(v):
        raise ValueError(f"'{v}' is not a valid CVSS3 vector.")
    return v


# cvss3 field
CVSS3Vector = Annotated[
    str,
    StringConstraints(
        strict=True,
        strip_whitespace=True,
    ),
    BeforeValidator(validate_with_is_cvss3),
]


def is_cvss4_valid(cvss_str: str) -> bool:
    """return True if cvss_str is a valid CVSS3 vector"""
    try:
        cvss.cvss4.CVSS4(cvss_str)
        return True

    except cvss.CVSSError:
        return False


def validate_with_is_cvss4(v: str) -> str:
    if not is_cvss4_valid(v):
        raise ValueError(f"'{v}' is not a valid CVSS3 vector.")
    return v


# cvss4 field
CVSS4Vector = Annotated[
    str,
    StringConstraints(
        strict=True,
        strip_whitespace=True,
    ),
    BeforeValidator(validate_with_is_cvss4),
]

# cve id field
CVEID = Annotated[
    str,
    StringConstraints(
        pattern=r"^CVE-\d{4}-\d{4,7}$",
        strict=True,
        strip_whitespace=True,
    ),
]

# create dynamic CVEID validator
cveid_validator = TypeAdapter(CVEID)



import pytest

from pydantic_core import ValidationError

from aegis_ai.agents import rh_feature_agent
from aegis_ai.data_models import CVEID
from aegis_ai.features import component, cve
from tests.utils.llm_cache import llm_cache_retrieve

pytestmark = pytest.mark.asyncio


async def test_suggest_impact_with_test_model():
    def feature():
        return cve.SuggestImpact(rh_feature_agent).exec("CVE-2025-0725")

    result = await llm_cache_retrieve(feature)
    suggestimpact = cve.data_models.SuggestImpactModel.model_validate_json(result)
    assert isinstance(suggestimpact, cve.data_models.SuggestImpactModel)
    assert suggestimpact.impact == "LOW"


async def test_suggest_cwe_with_test_model():
    def feature():
        return cve.SuggestCWE(rh_feature_agent).exec("CVE-2025-0725")

    result = await llm_cache_retrieve(feature)
    suggestcwe = cve.data_models.SuggestCWEModel.model_validate_json(result)
    assert isinstance(suggestcwe, cve.data_models.SuggestCWEModel)
    assert suggestcwe.cwe == ["CWE-190", "CWE-120"]


async def test_identify_pii_with_test_model():
    def feature():
        cve_id = CVEID(
            "CVE-2025-0725"
        )  # we can directly use custom fields though auto validation happens during feature input
        return cve.IdentifyPII(rh_feature_agent).exec(cve_id)

    result = await llm_cache_retrieve(feature)
    piireport = cve.data_models.PIIReportModel.model_validate_json(result)
    assert isinstance(piireport, cve.data_models.PIIReportModel)
    assert not piireport.contains_PII  # is false


async def test_rewrite_description_with_test_model():
    def feature():
        return cve.RewriteDescriptionText(rh_feature_agent).exec("CVE-2025-0725")

    result = await llm_cache_retrieve(feature)
    rewritedescription = cve.data_models.RewriteDescriptionModel.model_validate_json(
        result
    )
    assert isinstance(rewritedescription, cve.data_models.RewriteDescriptionModel)
    assert (
        rewritedescription.rewritten_title
        == "libcurl: integer overflow in zlib decompression"
    )


async def test_rewrite_statement_with_test_model():
    def feature():
        return cve.RewriteStatementText(rh_feature_agent).exec("CVE-2025-0725")

    result = await llm_cache_retrieve(feature)
    rewritestatement = cve.data_models.RewriteStatementModel.model_validate_json(result)
    assert isinstance(rewritestatement, cve.data_models.RewriteStatementModel)
    assert (
        rewritestatement.rewritten_statement
        == "A flaw was found in libcurl, where a buffer overflow could be triggered. This issue occurs when libcurl, using a vulnerable version of the zlib library (1.2.0.3 or older), performs automatic gzip decompression of content-encoded HTTP responses. A malicious server could exploit this by sending a specially crafted response, causing an integer overflow in zlib that leads to a buffer overflow in libcurl, resulting in a denial of service. Red Hat Enterprise Linux is not affected by this vulnerability as it does not ship the vulnerable version of zlib. However, some Red Hat products that bundle the affected library may be impacted."
    )


async def test_cvss_diff_explain_with_test_model():
    def feature():
        return cve.CVSSDiffExplainer(rh_feature_agent).exec("CVE-2025-0725")

    result = await llm_cache_retrieve(feature)
    cvssdiffexplain = cve.data_models.CVSSDiffExplainerModel.model_validate_json(result)
    assert isinstance(cvssdiffexplain, cve.data_models.CVSSDiffExplainerModel)
    assert (
        cvssdiffexplain.redhat_cvss3_vector
        == "CVSS:3.1/AV:L/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:L"
    )


async def test_component_intelligence_test_model():
    def feature():
        return component.ComponentIntelligence(rh_feature_agent).exec("curl")

    result = await llm_cache_retrieve(feature)
    componentintelligence = (
        component.data_models.ComponentIntelligenceModel.model_validate_json(result)
    )
    assert isinstance(
        componentintelligence, component.data_models.ComponentIntelligenceModel
    )
    assert componentintelligence.popularity_score == 1
    assert componentintelligence.confidence == 0.95


async def test_suggest_impact_with_bad_cve_test_model():
    def feature():
        return cve.SuggestImpact(rh_feature_agent).exec("BAD-CVE-ID")

    with pytest.raises(ValidationError) as excinfo:
        await feature()

    assert "String should match pattern" in str(excinfo)



import logging
import os
from dataclasses import dataclass
from typing import List

from pydantic import Field

from pydantic_ai import (
    RunContext,
    Tool,
)

import osidb_bindings

from aegis_ai.data_models import CVEID, cveid_validator
from aegis_ai.tools import BaseToolOutput

logger = logging.getLogger(__name__)

osidb_server_uri = os.getenv("AEGIS_OSIDB_SERVER_URL", "https://localhost:8000")
osidb_retrieve_embargoed = os.getenv(
    "AEGIS_OSIDB_RETRIEVE_EMBARGOED", "false"
).lower() in ("true", "1", "t", "y", "yes")


@dataclass
class OsidbDependencies:
    test = 1


class CVE(BaseToolOutput):
    """ """

    cve_id: CVEID = Field(
        ...,
        description="The unique Common Vulnerabilities and Exposures (CVE) identifier for the security flaw.",
    )

    title: str = Field(
        ...,  # Make it required
        description="CVE title.",
    )
    statement: str = Field(
        None,
        description="CVE statement.",
    )
    comment_zero: str = Field(None, description="CVE comment_zero.")
    comments: str = Field(
        None,
        description="all public comments.",
    )
    description: str = Field(None, description="CVE cve_description.")
    components: List = Field(..., description="list of components")
    references: List = Field(..., description="list of references")
    affects: List = Field(..., description="list of affects")
    cvss_scores: List = Field(..., description="list of cvss scores")


async def osidb_retrieve(cve_id: CVEID):
    logger.info(f"retrieving {cve_id} from osidb")
    validated_cve_id = cveid_validator.validate_python(cve_id)

    try:
        session = osidb_bindings.new_session(osidb_server_uri=osidb_server_uri)

        # Retrieval of embargoed flaws is disabled by default, to enable set env var `AEGIS_OSIDB_RETRIEVE_EMBARGOED`
        flaw = session.flaws.retrieve(
            id=validated_cve_id,
            include_fields="cve_id,title,cve_description,cvss_scores,statement,components,comments,comment_zero,affects,references,embargoed",
        )

        # This logic is about default constraining LLM access to embargo information ... for additional programmatic safety, user acl always
        # dictates if a user has access or not.
        if not osidb_retrieve_embargoed and flaw.embargoed:
            logger.info(
                f"retrieved {validated_cve_id} from osidb but it is under embargo and AEGIS_OSIDB_RETRIEVE_EMBARGOED is set 'false'."
            )
            return None

        logger.info(f"{validated_cve_id}:{flaw.title}")
        comments = ""
        for i, comment in enumerate(flaw.comments):
            if i >= 15:  # FIXME: remove limit of 15 comments
                break
            if not comment.is_private:
                comments += str(comment.text) + " "
        affects = []
        for affect in flaw.affects:
            affects.append(
                {
                    "affected": affect.affectedness,
                    "ps_module": affect.ps_module,
                    "ps_product": affect.ps_product,
                    "ps_component": affect.ps_component,
                    "impact": affect.impact,
                    "not_affected_justification": affect.not_affected_justification,
                    "delegated_not_affected_justification": affect.delegated_not_affected_justification,
                }
            )
        references = []
        for reference in flaw.references:
            if hasattr(reference, "url") and reference.url:
                references.append(
                    {
                        "url": reference.url,
                    }
                )
        cvss_scores = []
        for cvss_score in flaw.cvss_scores:
            cvss_scores.append(
                {
                    "issuer": cvss_score.issuer,
                    "vector": cvss_score.vector,
                }
            )
        return CVE(
            cve_id=flaw.cve_id,
            title=f"{flaw.title}",
            comment_zero=flaw.comment_zero,
            comments=f"{comments}",
            statement=f"{flaw.statement}",
            description=f"{flaw.cve_description}",
            components=flaw.components,
            references=references,
            affects=affects,
            cvss_scores=cvss_scores,
        )
    except Exception as e:
        logger.error(
            f"We encountered an error during OSIDB retrieval of {validated_cve_id}: {e}"
        )


@Tool
async def osidb_tool(ctx: RunContext[OsidbDependencies], cve_id: CVEID) -> CVE:
    """
    Searches OSIDB by cve_id performing a lookup on CVE entity in OSIDB and returns structured information about it.

    Args:
        ctx: The RunContext provided by the Pydantic-AI agent, containing dependencies.
        cve_lookup_input: An object containing validated CVE ID (e.g., CVE-2024-30941).

    Returns:
        CVE: A Pydantic model containing the CVE entity's cve_id, title, description, severity or an error message.
    """
    logger.debug(cve_id)
    return await osidb_retrieve(cve_id)



"""
Aegis MCP - register mcp here

"""

from pydantic_ai.mcp import MCPServerStdio

# mcp-nvd: query NIST National Vulnerability Database (NVD)
# https://github.com/marcoeg/mcp-nvd
#
# requires NVD_API_KEY=
nvd_server = MCPServerStdio(
    "uv",
    args=[
        "run",
        "mcp-nvd",
    ],
)



import pytest

from pydantic_evals import Case, Dataset
from pydantic_evals.evaluators import Evaluator, EvaluatorContext

from aegis_ai.agents import rh_feature_agent
from aegis_ai.data_models import CVEID
from aegis_ai.features.cve import SuggestImpact, SuggestImpactModel

from evals.features.common import (
    common_feature_evals,
    create_llm_judge,
    handle_eval_report,
)


# dict to convert "IMPORTANT" to 8.0 etc
# the following line is needed for ruff to accept the aligned comments
# fmt: off
NUM_BY_IMPACT = {
    "NONE": 0.0,        # 0
    "LOW": 2.0,         # 0..4
    "MODERATE": 5.5,    # 4..7
    "IMPORTANT": 8.0,   # 7..9
    "CRITICAL": 9.5,    # 9..10
}
# fmt: on


class SuggestImpactCase(Case):
    def __init__(self, cve_id, impact, cvss3_score):
        """cve_id given as CVE-YYYY-NUM is the flaw we query Impact for.
        impact is the expected impact as string. cvss3_score is the expected
        score specified as float."""
        super().__init__(
            name=f"suggest-impact-for-{cve_id}",
            inputs=cve_id,
            expected_output={"impact": impact, "cvss3_score": cvss3_score},
            metadata={"difficulty": "easy"},
        )


class SuggestImpactEvaluator(Evaluator[str, SuggestImpactModel]):
    def evaluate(self, ctx: EvaluatorContext[str, SuggestImpactModel]) -> float:
        """return score based on actual and expected results"""
        # compare actual and expected impact
        imp = NUM_BY_IMPACT[ctx.output.impact]
        imp_exp = NUM_BY_IMPACT[ctx.expected_output["impact"]]
        score = 1.0 - abs(imp - imp_exp) / 10.0

        try:
            # compare actual and expected cvss3_score
            cvss3 = float(ctx.output.cvss3_score)
            cvss3_exp = ctx.expected_output["cvss3_score"]
            score *= 1.0 - abs(cvss3 - cvss3_exp) / 10.0
        except ValueError:
            # the provided cvss3_score field is not a number
            score -= 1.0

        conf_diff = ctx.output.confidence - score
        if 0.0 < conf_diff:
            # penalize confident models providing (partially) wrong results
            score -= conf_diff
        else:
            # negligibly penalize models providing correct results but low confidence
            score += conf_diff / 4.0

        return score


async def suggest_impact(cve_id: CVEID) -> SuggestImpactModel:
    """use rh_feature_agent to suggest Impact for the given CVE"""
    feature = SuggestImpact(rh_feature_agent)
    result = await feature.exec(cve_id)
    return result.output


# test cases
cases = [
    SuggestImpactCase("CVE-2022-48701", "MODERATE", 4.9),
    SuggestImpactCase("CVE-2024-53232", "MODERATE", 4.4),
    SuggestImpactCase("CVE-2025-23395", "MODERATE", 6.8),
    SuggestImpactCase("CVE-2025-5399", "MODERATE", 4.3),
    # TODO: add more cases
]

# evaluators
evals = common_feature_evals + [
    SuggestImpactEvaluator(),
    create_llm_judge(
        rubric="explanation does not mention which Red Hat products are affected"
    ),
]

# needed for asyncio event loop
pytestmark = pytest.mark.asyncio(loop_scope="session")


async def test_eval_suggest_impact():
    """suggest_impact evaluation entry point"""
    dataset = Dataset(cases=cases, evaluators=evals)
    report = await dataset.evaluate(suggest_impact)
    handle_eval_report(report)



import logging

from aegis_ai.agents import kb_agent
from aegis_ai.kb import RagSystem, DocumentInput, FactInput, RAGQuery


async def main():
    # Initialize the RAG system
    rag_system = RagSystem(
        pg_connection_string="postgresql://postgres:password@localhost:5432/aegis",
        embedding_model_name="sentence-transformers/all-MiniLM-L6-v2",
    )
    await rag_system.initialize()

    try:
        # 1. Add documents
        doc_text = "Red Hat announced OpenShift 4.10 with enhanced security features and improved developer experience."
        doc_meta = {"title": "OpenShift 4.10 Announcement", "author": "Red Hat"}
        await rag_system.add_document_to_vector_store(
            DocumentInput(text=doc_text, metadata=doc_meta)
        )

        doc_text = "The capital of France is Paris. Paris is known for its Eiffel Tower and Louvre Museum."
        doc_meta = {
            "source": "Wikipedia",
            "category": "Geography",
            "original_text_length": len(doc_text),
        }
        await rag_system.add_document_to_vector_store(
            DocumentInput(text=doc_text, metadata=doc_meta)
        )

        doc_text = "Python is a high-level, interpreted programming language. It is widely used for web development, data analysis, AI, and more."
        doc_meta = {
            "source": "Programming Guide",
            "category": "Technology",
            "original_text_length": len(doc_text),
        }
        await rag_system.add_document_to_vector_store(
            DocumentInput(text=doc_text, metadata=doc_meta)
        )

        # 2. Add facts
        fact_text = "CVE-2023-1234 is a critical vulnerability affecting OpenSSL."
        fact_metadata = {"source": "NVD", "severity": "Critical"}
        await rag_system.add_fact_to_vector_store(
            FactInput(fact=fact_text, metadata=fact_metadata)
        )

        fact = "Mount Everest is the highest mountain in the world."
        await rag_system.add_fact_to_vector_store(
            FactInput(fact=fact, metadata={"source": "Factbook"})
        )
        fact = "The color of the sky is mostly blue."
        await rag_system.add_fact_to_vector_store(
            FactInput(fact=fact, metadata={"source": "Factbook"})
        )

        # 3. Perform RAG queries
        query_text = "What is the capital of France ?"
        rag_query = RAGQuery(
            query=query_text, top_k_documents=1, top_k_facts=1, additional_context=""
        )

        rag_response = await rag_system.perform_rag_query(rag_query, kb_agent)
        print("---------------------")
        print(query_text)
        print("---------------------")
        print(rag_response)
        print("---------------------")

    finally:
        await rag_system.shutdown()


if __name__ == "__main__":
    import asyncio

    logging.basicConfig(level=logging.INFO)
    asyncio.run(main())



import asyncio
import inspect
import logging
import os
from pathlib import Path

logger = logging.getLogger(__name__)

# directory where we cache CVE data retrieved from OSIDB
LLM_CACHE_DIR = os.getenv("TEST_LLM_CACHE_DIR", "tests/llm_cache")

# global mutex for access to LLM_CACHE_DIR
# Note that cache hits (which is the most common case) are handle very quickly.
# So there is no need to implement any per-file locking for the OSIDB cache.
cache_lock = asyncio.Lock()


async def llm_cache_retrieve(feature):
    """Return cached LLM data if available.  If not, retrieve LLM data"""

    # use test function name as name of related cache file
    test_name = inspect.stack()[1].function
    cache_file = Path(LLM_CACHE_DIR, f"{test_name}.json")

    # acquire global mutex to access LLM_CACHE_DIR
    async with cache_lock:
        try:
            # check whether the LLM data is cached already
            with open(cache_file, "r") as f:
                json_data = f.read()

            # try to load data from the existing JSON file
            data = json_data
            logger.info(f'read LLM data from "{cache_file}"')

        except OSError:
            # cached LLM data not available -> query LLM
            llm_result = await feature()
            data = llm_result.output
            logger.info(data)
            logger.info(f'writing LLM data cache to "{cache_file}"')
            with open(cache_file, "w") as f:
                f.write(data.model_dump_json(indent=4))
                f.write("\n")
    return data



import logging

from aegis_ai.data_models import CVEID
from aegis_ai.features import Feature
from aegis_ai.features.cve.data_models import (
    CVSSDiffExplainerModel,
    SuggestImpactModel,
    SuggestCWEModel,
    PIIReportModel,
    RewriteStatementModel,
    RewriteDescriptionModel,
)
from aegis_ai.features.cve.data_models import CVEFeatureInput
from aegis_ai.prompt import AegisPrompt

logger = logging.getLogger(__name__)


class SuggestImpact(Feature):
    """Based on current CVE information and context assert an aggregated impact."""

    async def exec(self, cve_id: CVEID):
        prompt = AegisPrompt(
            user_instruction="Your task is to meticulously examine the provided CVE JSON object and suggest an overall impact rating for the CVE",
            goals="""
                # Task:
                Given a CVE ID or CVE description containing a description of a vulnerability, draft CVSS, affected components and other information generate an impact rating based on the following four-point scale used by Red Hat:
                * CRITICAL: This rating is given to flaws that could be easily exploited by a remote unauthenticated attacker and lead to system compromise (arbitrary code execution) without requiring user interaction, or easily cause system compromise via inference end points on AI systems. Flaws that require authentication, local or physical access to a system, or an unlikely configuration are not classified as Critical impact. These are the types of vulnerabilities that can be exploited by worms.
                * IMPORTANT: This rating is given to flaws that can easily compromise the confidentiality, integrity or availability of resources. These are the types of vulnerabilities that allow local or authenticated users to gain additional privileges, allow unauthenticated remote users to view resources that should otherwise be protected by authentication or other controls, allow authenticated remote users to execute arbitrary code, allow remote users to cause a denial of service, or can cause system compromise via inference end points on AI systems.
                * MODERATE: This rating is given to flaws that may be more difficult to exploit but could still lead to some compromise of the confidentiality, integrity or availability of resources under certain circumstances. It is also given to flaws that could be exploited to cause denial of service-like conditions on AI systems via an inference end point, or allow attackers to steal other users’ data from the AI system without authorization. These are the types of vulnerabilities that could have had a Critical or Important impact but are less easily exploited based on a technical evaluation of the flaw and/or affect unlikely configurations.
                * LOW: This rating is given to all other issues that may have a security impact. These are the types of vulnerabilities that are believed to require unlikely circumstances to be able to be exploited, or where a successful exploit would give minimal consequences. This includes flaws that are present in a program’s source code but to which no current or theoretically possible, but unproven, exploitation vectors exist or were found during the technical analysis of the flaw.
                
                # Instructions for Analysis:
                
                Thorough Traversal: Recursively traverse the entire JSON object, including nested arrays and objects.
            """,
            rules="""
                1.  Analyze the provided CVE data for information related to:
                    * Attack vector (remote, local, physical)
                    * Authentication requirements
                    * User interaction requirements
                    * Impact on confidentiality
                    * Impact on integrity
                    * Impact on availability
                    * Potential for arbitrary code execution
                    * Potential for privilege escalation
                    * Potential for denial of service
                  and generate a red hat specific CVSS3 and CVSS4 score to help suggest impact.
                2. Analysis should be based on the provided JSON against all known CVEs affecting Red Hat products.
                3. The reasoning should not take into account which Red Hat products are affected by the flaw.
                4. Denial of Service (DoS) flaws are usually not IMPORTANT if their scope is limited to an application.
                5. A User Interaction is usually Required (UI:R in CVSS) in case an application connects a malicious server to trigger the flaw.
                6. Provide a confidence % in how accurate (based on training material, reasoning) this assessment is.
            """,
            context=CVEFeatureInput(cve_id=cve_id),
            output_schema=SuggestImpactModel.model_json_schema(),
        )
        return await self.agent.run(prompt.to_string(), output_type=SuggestImpactModel)


class SuggestCWE(Feature):
    """Based on current CVE information and context assert CWE(s)."""

    async def exec(self, cve_id: CVEID):
        prompt = AegisPrompt(
            user_instruction="Your task is to meticulously examine the provided CVE JSON object and suggest CWE-ID(s) for the CVE",
            goals="""
                Given a description of a Common Vulnerabilities and Exposures (CVE), accurately predict its Common Weakness Enumeration (CWE). List all CWE that are applicable.
                
                Provide the predicted CWE identifier and a brief explanation for the reasoning behind the prediction.
                
                Provide a confidence % representing how confident you are this is correct CWE.
                
                Assist users in understanding the potential software weakness associated with a given vulnerability.
                
                # Instructions for Analysis:
                
                Thorough Traversal: Recursively traverse the entire provided JSON object, including nested arrays and objects.
            """,
            rules="""
                Input Processing:
                
                a) Receive and process textual descriptions of CVEs.
                
                b) Identify key characteristics and patterns within the CVE description relevant to potential software weaknesses.
                
                # CWE Prediction:
                
                a) Based on the analysis of the CVE description, predict the most likely CWE.
                
                b) Provide the standard CWE identifier (e.g., CWE-119).
                
                c) Offer a concise explanation outlining the connection between the CVE description and the predicted CWE.
                
                d) Avoid predicting CWEs that are discouraged or prohibited for Vulnerability Mapping by MITRE.  In particular, do not suggest CWE-264 and CWE-269.
            """,
            context=CVEFeatureInput(cve_id=cve_id),
            output_schema=SuggestCWEModel.model_json_schema(),
        )
        return await self.agent.run(prompt.to_string(), output_type=SuggestCWEModel)


class IdentifyPII(Feature):
    """Based on current CVE information (public comments, description, statement) and context assert if it contains any PII."""

    async def exec(self, cve_id: CVEID):
        prompt = AegisPrompt(
            user_instruction="Your task is to meticulously examine the provided CVE JSON object and identify any instances of Personally Identifiable Information (PII).",
            goals="""
            # Definition of PII for this task:
            
                PII includes, but is not limited to:
                    Direct Identifiers: Full names, email addresses, phone numbers, passwords, social security numbers (SSN), national identification numbers, passport numbers, driver's license numbers, bank account numbers, credit card numbers.
                    Indirect Identifiers (that can be linked to an individual): Dates of birth, home addresses, precise geographical coordinates, IP addresses, MAC addresses, device IDs, biometric data (e.g., fingerprints, facial recognition data), unique identifiers from cookies or advertising.
                    Sensitive Information: Health information, genetic information, racial or ethnic origin, political opinions, religious or philosophical beliefs, trade union membership, sexual orientation, criminal records.
            
            # Instructions for Analysis:
            
                Thorough Traversal: Recursively traverse the entire JSON object, including nested arrays and objects.
                Key and Value Analysis: Examine both the keys (field names) and values within the JSON. PII can sometimes be indicated by a key name (e.g., "email", "phone_number") even if the value is empty or generic, but focus primarily on values.
                Pattern Recognition: Look for common patterns associated with PII, such as:
                    Email address formats (e.g., user@domain.com)
                    Phone number formats (various international formats)
                    Numerical sequences that resemble Social Security numbers/national IDs/credit card numbers (consider common lengths and checksums if possible, but err on the side of caution).
                    Keywords in keys or values that suggest PII (e.g., "name", "address", "dob", "ssn", "health", "policy").
                    Anything that looks like a secure password or secret
                Contextual Understanding: Consider the context of the data. For example, a street name alone isn't PII, but a combination of street, city, and postal code often is.
            
                The following example json
            
                ```json
                {"title":"this contains jim@webomposite.com"}
                ```
                So the analysis would identify `jim@webcomposite.com` as PII because it contains an email address.
            
                Taking another example
                ```json
                {"title":"this contains 035-48-2559"},"description":"The fone is +420733228297"}
                ```
                The analysis should identify `035-48-3559` as a ssn and identify +420733228297 as a phone number.
            
                Example 1: Name and Home Address
                ```json
                {
                  "user_profile": {
                    "name": "Alice Wonderland",
                    "contact": {
                      "street": "123 Rabbit Hole Lane",
                      "city": "Wonderland",
                      "zip": "10001"
                    },
                    "preferences": ["tea party", "chess"]
                  },
                  "activity_log": {
                    "event": "logged_in",
                    "timestamp": "2024-05-29T10:30:00Z"
                  }
                }
                ```
            
                Analysis Should Identify addresses:
            
                    Alice Wonderland as a Full Name.
                    123 Rabbit Hole Lane as part of a Home Address.
                    Wonderland as part of a Home Address.
                    10001 as part of a Home Address.
            
                Example 2: Date of Birth and IP Address
                ```json
                {
                  "system_event": {
                    "event_id": "SYS-87654",
                    "client_ip": "192.168.1.100",
                    "details": {
                      "user_agent": "Mozilla/5.0...",
                      "session_start": "2025-05-29T14:15:00Z"
                    }
                  },
                  "customer_data": {
                    "customer_id": "CUST-9988",
                    "dob": "1990-07-15"
                  }
                }
                ```
                Analysis Should Identify:
            
                    192.168.1.100 as an IP Address.
                    1990-07-15 as a Date of Birth.
            
                Example 3: Credit Card Number and Health Information
                ```json
                {
                  "payment_info": {
                    "order_id": "ORD-54321",
                    "card_number": "4111222233334444",
                    "expiry_date": "12/28"
                  },
                  "medical_record": {
                    "record_id": "MR-ALPHA-001",
                    "diagnosis": "Seasonal allergies",
                    "patient_notes": "Patient reported sneezing and watery eyes.",
                    "genetic_marker": "HLA-DRB1"
                  }
                }
                ```
                Analysis Should Identify:
            
                    4111222233334444 as a Credit Card Number.
                    Seasonal allergies as Health Information.
                    Patient reported sneezing and watery eyes. as Health Information.
                    HLA-DRB1 as Genetic Information.
            
                Example 4: National Identification Number and Biometric Data Hint
                ```json
                {
                  "enrollment_form": {
                    "form_id": "ENR-007",
                    "applicant_id": "ID-9876543210",
                    "biometric_hash": "fingerprint_hash_abc123xyz",
                    "nationality": "Czech"
                  }
                }
            
                Analysis Should Identify:
            
                    ID-9876543210 as a National Identification Number (assuming context where ID- prefix implies a national ID).
                    fingerprint_hash_abc123xyz as Biometric Data.
            """,
            rules="""
            # Rules for output fields
                explanation:
                    If PII is found, create a bulleted list where each item is formatted as PII type:"exact string".
                    The exact string cannot be empty.
                    The PII type should be a concise description (e.g., "Gender","Race","Email Address", "Phone Number", "Physical Address", "Health Information").
                    The "exact string" must be the literal value from the JSON that constitutes the PII.
                    A separate bullet point is used for each instance of PII found.
                    If NO PII is found, this field should be an empty string ("").
                confidence: Generate a score between 0.00 and 1.00 (two decimal places) indicating confidence in the PII analysis. A higher score means greater certainty.
                contains_PII: Set to true if any PII was identified, false otherwise.
            
            # Additional guidelines for report
            - Only report on PII that is present in the provided json.
            - Explanation should only contain list of PII found.
            - Do not include any code in this report.
            - Do not gratitiously repeat information.
            - Do not insert line breaks (ex. \n).
            - Do not invent or add any information that is not present in the given context.
            
            If PII is found, include a bulleted list of each occurrence of PII in explanation.
            Each bullet point should contain the PII type and the exact string found, formatted as "PII type:'exact string'".
            If no PII is found, leave this section empty.
            
            To determine how to set contains_PII:
            1. If you find at least one instance of PII, set contains_pii:True.
            2. If you do not find any PII, set contains_pii:False.

            """,
            context=CVEFeatureInput(cve_id=cve_id),
            output_schema=PIIReportModel.model_json_schema(),
        )
        return await self.agent.run(prompt.to_string(), output_type=PIIReportModel)


class RewriteDescriptionText(Feature):
    """Based on current CVE information and context rewrite/create description and title."""

    async def exec(self, cve_id: CVEID):
        prompt = AegisPrompt(
            user_instruction="Your task is to meticulously examine the provided JSON object and rewrite cve description for it. The goal of the description is to briefly provide an overview of the CVE. If the cve description exists, rewrite it - if it does not exist suggest new text.",
            goals="""
                * Evaluate the quality of CVE description text from the perspective of a security analyst.
                * Provide an overall text quality score for a given CVE description based on a comparison with previous Red Hat CVEs.
                * Offer a confidence score for the analysis performed.
                * Generate an improved, rewritten version of the CVE description for clarity.
            """,
            rules="""
                1) Input Analysis:
                
                a) When provided with a CVE identifier or a CVE description, analyze the text for clarity, conciseness, and completeness from a security analyst's viewpoint.
                
                b) Identify any jargon, ambiguity, or missing information that could hinder understanding.
                
                c) Ensure writing is from the point of view of Red Hat security analysis
                
                2) Scoring:
                
                a) Calculate a text quality score reflecting how well the CVE description explains the security vulnerability, drawing comparisons with historical Red Hat CVE descriptions as a benchmark.
                
                b) Assign a confidence score to the analysis, indicating the certainty of the evaluation and the quality score.
                
                3) Rewriting:
                
                a) Generate an alternative CVE description that is clearer, more concise, and easier for a security analyst to understand.
                
                b) Ensure the rewritten description accurately conveys the essential information about the vulnerability.
                
                c) Maintain a professional and informative tone in the rewritten description.
                
                d) Generate only the vulnerability description in the following format:\n"
                
                   "A flaw was found in [product/component]. This vulnerability allows [impact description] via [attack vector].\n"
                
                   Do not include any versioning information, introductory text, or explanations. Return only the description.
                
                e) Generate a short and precise title for this vulnerability.
                   Avoid including versioning, introductory text, or explanations.
                   The title should be concise, professional, and no longer than 7 words.
                   The title should contain the product name and the type of vulnerability.
                
                # Further Guidelines for writing CVE description
                
                The CVE Description field should provide information to help customers quickly and easily understand the flaw's threat
                to their system.
                
                CVE Description is a basic description of the following information:
                
                    * The type of vulnerability and where it exists
                    * Who or what can exploit the flaw
                    * How the flaw can be exploited
                
                The following is a guideline for the Description structure:
                
                The type of flaw (based on CWE) and what it can do
                For example, "A buffer overflow", "An integer overflow", "A NULL pointer dereference", or "A privilege escalation flaw". Suggested phrasing examples:
                    "A flawed bounds check in the xxxx function leads to... "
                    "A Cross-Site Request Forgery (CSRF) issue can lead to..."
                    "A cross-site scripting (XSS) flaw leading to..."
                Who can exploit the flaw
                For example, local or remote attacker, authenticated or unauthenticated user or attacker, privileged or unprivileged guest user, man-in-the-middle attacker, or a malicious server. Suggested phrasing examples:
                    "An attacker with CREATE and DROP table privileges and shell access to the database server could use this issue to ..."
                    "A local attacker could use this issue to cause a denial of service by mounting a specially crafted [file system type, such as ext4] file system."
                    "This flaw allows an attacker to..."
                How the flaw can be exploited
                For example, "specially crafted packet or request" or "specially crafted text file". Suggested phrasing examples:
                    "An attacker could create a specially crafted image file that, when opened by a victim, could cause an application to crash or allow arbitrary code execution."
                    "If a carefully crafted file-type file was loaded by an application linked against [affected-library-name], the application could crash or allow arbitrary code execution with the privileges of the user running the application."
                Exploitation consequences
                For example, a denial of service, code execution, privilege escalation, or information disclosure. Suggested phrasing examples:
                    "A user in a guest could leverage this flaw to cause a denial of service (guest hang or crash) or possibly escalate their privileges within the host."
                    "This could lead to a denial of service if a user browsed a specially crafted [file system type, such as ext4] file system, for example, by running 'ls'."
                
                ## Things to do:
                
                The description should be unique and distinguishable between similar-looking vulnerabilities. For example, if there are two CVEs reported for a component and both of them are CSRF attacks, ensure customers can differentiate one from another based on attack vector, attack surface, or data breach.
                Follow the correct naming conventions and capitalization wherever possible, such as product names. For example, using names such as Gluster or Samba, and using protocols such as HTTPS, FTP, or TLS.
                Use the Statement to anticipate customer questions. In particular, if our flaw impact (Severity rating) is significantly different from the NVD severity rating defined by their CVSS Base Score, this should be explained. For instance, if the NVD CVSS Base Score is in the range they define as High or Critical, but our rating is Moderate or Low, customers will seek an explanation for why our rating is lower. This explanation is also important for customers responding to reports from vulnerability scanners.
                Severity ratings and CVSS Base Score Ranges:
                    none    0.0
                    low       0.1 - 3.9
                    medium  4.0 - 6.9
                    high     7.0 - 8.9
                    critical 9.0 - 10.0
                
                ## Things not to do:
                
                The use of acronyms should be avoided when possible because customers might not know the acronym. When it’s necessary to use an acronym, add the corresponding full form. For example, use "Active Directory Domain Controller (AD DC)" rather than just writing “AD DC”.
                Engineers should refrain from using generic Confidentiality, Integrity, and Availability (CIA) statement boilerplate templates. For example, "The highest threat from this vulnerability is to data confidentiality". Instead, describe what kind of restricted information can be obtained and the amount of information disclosure.
                Information that is already included in other fields. For example, the CVE number or affected versions, which belong in Comment #0 of the flaw bug.
                
                If multiple Red Hat products are affected, do not include the name of a specific product in the description. The per-product technical description, affected state, or any other special cases belongs in the Statement.
                
                # Further Guidelines for CVE Title
                
                The title should describe *only* the core technical issue based *only* on the provided input.
                
                ### Instructions:
                
                1.  Start the paragraph *exactly* like this: "A flaw was found in 'component_name'."
                2.  Continue by *briefly* summarizing from the "Original Description Snippet" and other hints:
                    * The **technical flaw** and its direct effect (e.g., 'buffer overflow leads to...').
                    * The **attacker profile** (e.g., 'local attacker can...').
                    * The **exploitation vector** (e.g., '...via a crafted file.').
                    * The **primary consequence** (e.g., 'This can cause a denial of service.').
                3.  Combine these points into a single, flowing paragraph.
                4.  **Crucially, DO NOT include:**
                    * Information about disclosure status.
                    * Any classification (like 'problematic', 'important').
                    * How the bug was found (like 'fuzzer-identified').
                    * Plans for fixing it or mitigation efforts (these are handled separately).
                    * CVSS scores or specific vector strings (these are displayed separately).
                    * Any generic text, explanations, or meta-commentary not directly describing the flaw's technical nature.
                5.  **DO NOT** use headings, bold text (other than the starting phrase if needed by a template), bullet points, or line breaks. It must be one single paragraph.
            """,
            context=CVEFeatureInput(cve_id=cve_id),
            output_schema=RewriteDescriptionModel.model_json_schema(),
        )
        return await self.agent.run(
            prompt.to_string(), output_type=RewriteDescriptionModel
        )


class RewriteStatementText(Feature):
    """Based on current CVE information and context rewrite/create statement."""

    async def exec(self, cve_id: CVEID):
        prompt = AegisPrompt(
            user_instruction="Your task is to meticulously examine the provided JSON object and rewrite CVE statement for it. The goal of the statement is explain the context for the CVE impact with respect to Red Hat supported products. If the CVE statement exists, rewrite it - if it does not exist suggest new text.",
            goals="""
                * Evaluate the quality of existing CVE statement text from the perspective of a security analyst.
                * Provide an overall text quality score for a given CVE description based on a comparison with previous Red Hat CVEs.
                * Offer a confidence score for the analysis performed.
                * Generate an improved, rewritten version of the CVE statement for clarity.
                * The rewritten statement does not duplicate information provided in the description field.
                * The rewritten statement does not include any low-level technical details, such as specific code changes.
                * The rewritten statement does not suggest customers to apply upstream patches or to rebuild software from source code.
                * If no additional context is available, the rewritten statement is empty.
            """,
            rules="""
                1) Input Analysis:
                
                When provided with a CVE identifier or a CVE description, analyze the CVE statement text for clarity, conciseness, and correctness from a security analyst's viewpoint.
                
                The CVE statement will explain why an impact is either higher or lower than expected, especially if Red Hat products are not impacted.
                
                2) Scoring:
                
                a) Calculate a text quality score reflecting how well the CVE statement explains the security vulnerability, drawing comparisons with historical Red Hat CVE statements as a benchmark.
                
                b) Assign a confidence score to the analysis, indicating the certainty of the evaluation and the quality score.
                
                3) Rewriting:
                
                a) Generate an alternative CVE statement that is clearer, more concise, and easier for customers to understand.
                
                b) Ensure the rewritten statement accurately conveys the context and rationale for the impact score modulo relation to Red Hat
                 supported products.
                
                c) Do not duplicate information available in other fields, such as flaw description.  Do not describe low-level technical details.
                
                
                # Further Guidelines to Writing a Comprehensive Vulnerability Statement
                
                A CVE's statement covers more than just the rationale behind an impact rating. Additionally it touches on:
                
                  - A direct opportunity to provide value beyond what’s easily found in an internet search.
                  - Additional technical details which clarify the scope of impact. Make it easy for the subset of unaffected customers (because of how they configure their deployment) to discover this.
                  - Explaining the affected function at code level is too much detail, it is better to explain:
                      - how a vulnerable function/service impacts the system
                      - what successful exploitation looks like
                      - or how difficult it is to exploit
                
                A flaw's Statement provides additional information about Red Hat’s response to a security issue and displays prominently on the flaw's CVE page where customers are likely to see it.
                
                Statements are especially important for flaws with a higher impact, which are likely to receive more customer attention because they hit the media, are in a critical component, or are more highly rated by other sources. Statements are even more important when our triage decisions may seem surprising from a naïve reading of description & CVSS.
                
                If a Statement prevents a support call, it has succeeded. If secalert is asked to help a customer understand a flaw, the answer can often be summarized in a Statement, and the customer or support person is asked to confirm whether the Statement answers their question. This interaction can then inform future Statements.
                
                ## Elements of an excellent statement
                
                To write an effective and excellent statement, several elements are necessary. The elements provide specific information that customers will find useful.
                
                  - Explains the RH impact rating and any difference between industry ratings utilizing CIA logic.
                  - Explains any differences in CVSS vectors in a Red Hat calculated score.
                  - References affected Red Hat components (vs upstream).
                  - Includes conditions and technical detail specific to Red Hat configuration, patching, and build details.
                  - Includes any statements of fact as necessary.
                
                ## Language and way of speaking
                
                Do not worry about the complexity of language. The statement should not contain complicated, legal sounding content. Keep the statements conversational, as if one engineer is talking to another. This will make the information easier to follow and understand.
                
                  - Avoid overly complicated language.
                  - Write simply and clearly, using words that everyone can understand.
                  - Sound confident.
            """,
            context=CVEFeatureInput(cve_id=cve_id),
            output_schema=RewriteStatementModel.model_json_schema(),
        )
        return await self.agent.run(
            prompt.to_string(), output_type=RewriteStatementModel
        )


class CVSSDiffExplainer(Feature):
    """Based on current CVE information and context explain CVSS score diff between nvd and rh."""

    async def exec(self, cve_id: CVEID):
        prompt = AegisPrompt(
            user_instruction="Explain the differences of cvss score attributed to CVE between supplied redhat CVE context and nvd CVE.",
            goals="""
                * Given a Common Vulnerabilities and Exposures (CVE) identifier, retrieve  its Red Hat CVSS and NVD(NIST) CVSS score
                * compare the Red Hat Common Vulnerability Scoring System (CVSS) score with the CVSS score reported by the National Vulnerability Database (NVD).
                * Identify and clearly explain any differences between the Red Hat and NVD CVSS scores.
                * Provide context for the CVSS metrics and how they contribute to the overall score, focusing on explaining why the scores might differ.
            """,
            # FIXME: Do we ever track temporal and environmental scores in OSIM?
            # I have never seen these metrics included in the cvss3/cvss4 fields.
            rules="""
                1) Input Analysis:
                
                    a) Accept a Red Hat CVE identifier as the primary input.
                
                    b) Validate the format of the CVE identifier.
                
                2) Data Retrieval:
                
                    a) Access or simulate access to information regarding the CVSS scores for the given CVE from both Red Hat and NVD sources.
                
                    b) Clearly state the retrieved CVSS base scores, temporal scores (if available), and environmental scores (if available) from both sources.
                
                    c) Include the CVSS vector strings for both Red Hat and NVD.
                
                3) Comparison and Explanation:
                
                    a) Directly compare the base, temporal, and environmental scores (if present) from both sources.
                
                    b) Identify specific differences in the individual CVSS metrics (e.g., Attack Vector, Attack Complexity, Confidentiality Impact) that contribute to any overall score discrepancies.
                
                    c) Explain the potential reasons for these differences, such as variations in vulnerability assessment, data interpretation, or the timing of the analysis by each organization.
                
                    d) Use clear and concise language, avoiding technical jargon where possible, or explaining it when necessary.
                
                    e) If there is no difference in the CVSS scores, or the information is not available, return an empty explanation.
                
                
                Format
                If there is no difference in the CVSS scores, provide an empty explanation.  Otherwise:
                * first line should contain Red Hat CVSS
                * second line should contain NVD CVSS
                * third line should contain confidence % in explaining the difference
                * Next provide rationale and explain difference between Red Hat and NVD CVSS score
                
                
                # JSON for analysis
                The cvss diff analysis performed on supplied CVE context which means you also must retrieve NVD (NIST) cvss for this CVE
                and compare and explain why there is a difference.
            """,
            context=CVEFeatureInput(cve_id=cve_id),
            output_schema=CVSSDiffExplainerModel.model_json_schema(),
        )
        return await self.agent.run(
            prompt.to_string(), output_type=CVSSDiffExplainerModel
        )



"""
Aegis chat: example app showing web chat.

"""

from __future__ import annotations as _annotations

from pathlib import Path

import json
import logging
from datetime import datetime, timezone
from typing import Annotated, List, Literal, Optional

import fastapi
from fastapi import Form
from fastapi.responses import FileResponse, StreamingResponse
from typing_extensions import TypedDict

from pydantic_ai.exceptions import UnexpectedModelBehavior
from pydantic_ai.messages import (
    ModelMessage,
    ModelRequest,
    ModelResponse,
    TextPart,
    UserPromptPart,
)
from pydantic import BaseModel

from aegis_ai.agents import public_feature_agent

logger = logging.getLogger(__name__)
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)


class AgentOutputModel(BaseModel):
    text: str = ""  # Default for simple streaming


THIS_DIR = Path(__file__).parent

app = fastapi.FastAPI()


@app.get("/")
async def index() -> FileResponse:
    return FileResponse((THIS_DIR / "chat_app.html"), media_type="text/html")


@app.get("/chat_app.js")
async def main_js() -> FileResponse:
    return FileResponse((THIS_DIR / "chat_app.js"), media_type="text/javascript")


@app.get("/logo.png")
async def logo_png() -> FileResponse:
    return FileResponse((THIS_DIR / "logo.png"), media_type="image/png")


class ChatMessage(TypedDict):
    role: Literal["user", "model"]
    timestamp: str
    content: str


def to_chat_message(m: ModelMessage) -> ChatMessage:
    first_part = m.parts[0]
    if isinstance(m, ModelRequest):
        if isinstance(first_part, UserPromptPart):
            assert isinstance(first_part.content, str)
            return {
                "role": "user",
                "timestamp": first_part.timestamp.isoformat(),
                "content": first_part.content,
            }
    elif isinstance(m, ModelResponse):
        if isinstance(first_part, TextPart):
            # Use the timestamp from the ModelResponse itself if available,
            # or from the first part. ModelResponse has a timestamp field.
            return {
                "role": "model",
                "timestamp": m.timestamp.isoformat(),  # ModelResponse has timestamp
                "content": first_part.content,
            }
    raise UnexpectedModelBehavior(f"Unexpected message type for chat app: {m}")


@app.post("/chat/")
async def post_chat(
    prompt: Annotated[str, Form()],
    history_json: Annotated[Optional[str], Form()] = None,
) -> StreamingResponse:
    async def stream_messages():
        user_chat_message = {
            "system_prompt": "You are a helpful assistant with access to several tools. You should use these tools to answer questions that require external information.**Prioritize using tools before trying to answer directly.**",
            "role": "user",
            "timestamp": datetime.now(tz=timezone.utc).isoformat(),
            "content": prompt,
        }
        yield json.dumps(user_chat_message).encode("utf-8") + b"\n"

        message_history: List[ModelMessage] = []
        if history_json:
            try:
                client_history: List[ChatMessage] = json.loads(history_json)
                for msg in client_history:
                    if msg["role"] == "user":
                        # PASS TIMESTAMP TO UserPromptPart
                        message_history.append(
                            ModelRequest(parts=[UserPromptPart(content=msg["content"])])
                        )
                    elif msg["role"] == "model":
                        # PASS TIMESTAMP TO TextPart
                        message_history.append(
                            ModelResponse(parts=[TextPart(content=msg["content"])])
                        )
            except (json.JSONDecodeError, KeyError, ValueError) as e:
                logger.error(
                    f"Failed to parse history_json from client: {e}. Received: {history_json[:200]}..."
                )
            except Exception as e:
                logger.error(f"Unexpected error parsing history_json: {e}")

        # TODO: .run_stream does not invoke tools use Agent.iter instead
        async with public_feature_agent.run_stream(
            prompt, message_history=message_history
        ) as result:
            ai_response_content = ""

            async for text_part_content in result.stream(debounce_by=0.01):
                m = ModelResponse(
                    parts=[TextPart(content=text_part_content)],
                    timestamp=datetime.now(tz=timezone.utc),
                )  # Use current time for display
                yield json.dumps(to_chat_message(m)).encode("utf-8") + b"\n"
                ai_response_content += text_part_content

    return StreamingResponse(stream_messages(), media_type="text/plain")
